package DiffAlgebra;

# # diffAlg - Differential algebra library
# # 
# # Copyright (C) 2012 Manuel Dubinsky, Cesar Massari, Ariel Molinuevo
# # 
# # 
# # This program is free software: you can redistribute it and/or modify
# # it under the terms of the GNU General Public License as published by
# # the Free Software Foundation, either version 3 of the License, or
# # (at your option) any later version.
# # This program is distributed in the hope that it will be useful,
# # but WITHOUT ANY WARRANTY; without even the implied warranty of
# # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# # GNU General Public License for more details.
# # You should have received a copy of the GNU General Public License
# # along with this program. If not, see <http://www.gnu.org/licenses/>.


# #!/usr/bin/perl

# use warnings ;
# use strict ;
use DB_File ;

my @n_upla;

sub combN_R_Internal {
	my ($FH, $pos, $n, $r, $lenTupla, @nrTop) = @_;

	for (my $i = $pos; $i <= $n - $r + 1; $i++) {
		$n_upla[$lenTupla-$r] = $i;
		if ($r == 1) {
			my $value;
			if (@nrTop) {
				$value = join(",", @nrTop[@n_upla]);
			} else {
				$value = join(",", @n_upla);
			}
			$$FH{$value} = undef;
		}
		if ($r > 1) {
			combN_R_Internal($FH, $i+1, $n, $r - 1, $lenTupla, @nrTop);
		}
	}
}

sub combN_R {
	my ($FH, $pos, $n, $r, $lenTupla, @nrTop) = @_;
	@n_upla = ();
	if (@nrTop) {
		combN_R_Internal($FH, $pos, scalar(@nrTop), $r, $lenTupla, (-1,@nrTop));
	} else {
		combN_R_Internal(@_);
	}
}

sub monomialsDegrees_Internal {
	my ($FH, $pos, $vars, $degree, @ndTop) = @_;

	my $maxValue = !@ndTop || $degree <= $ndTop[$pos-1] ? $degree : $ndTop[$pos-1];
	if ($pos < $vars) {
		for (my $i = $maxValue; $i >= 0; $i--) {
			$n_upla[$pos-1] = $i;
			monomialsDegrees_Internal($FH, $pos+1, $vars, $degree-$i, @ndTop);
		}
	} elsif ($pos == $vars) {
		# NOTA: puede pasar que no se pueda completar el grado en la ultima posicion...
		if (!@ndTop || $degree <= $ndTop[$pos-1]) {
			$n_upla[$pos-1] = $maxValue;
			my $value = join(",", @n_upla);
			$$FH{$value} = undef;
		}
	}
}

sub monomialsDegrees {
	@n_upla = ();
	monomialsDegrees_Internal(@_);
}

sub formCoefficients {
	my ($n, $r, $d, $coefFile, $varName) = @_;
# $n: cantidad de vars $r: r-forma $d: grado de los polinomios
	my $formsFile =  "tmp/" . $n . "_" . $r . ".nr";	
	my %h;
	if (! -e $formsFile) {		
		tie %h, "DB_File", $formsFile or die "Cannot open file $formsFile: $!\n";
		combN_R(\%h, 1, $n, $r, $r);
		untie %h;
	}
	
	my $monomialsFile = "tmp/" . $n . "_" . $d . ".nd";
	if (! -e $monomialsFile) {
		tie %h, "DB_File", $monomialsFile or die "Cannot open file $monomialsFile: $!\n";
		monomialsDegrees(\%h, 1, $n, $d);
		untie %h;
	}

	if (-e $coefFile) {
		unlink($coefFile);
	}
	my (%FH_COEF, %FH_FORMS, %FH_MONOM);
	tie %FH_COEF, "DB_File", $coefFile or die "Cannot open file $coefFile: $!\n";
	tie %FH_FORMS, "DB_File", $formsFile or die "Cannot open file 1: $!\n";
	tie %FH_MONOM, "DB_File", $monomialsFile or die "Cannot open file 2: $!\n";
	map {
		my $form = $_;
		map {
			my $value = $form . ";" . $_;
			$FH_COEF{$value} = !$varName ? "" : $varName . "_" . $value;
		} keys(%FH_MONOM);
	} keys(%FH_FORMS);
	untie %FH_COEF;
	untie %FH_FORMS;
	untie %FH_MONOM;
	
	return $coefFile;
}

sub formKey {
	my ($d, $coefs) = @_;
	return join(";", (join(",", @$d), join(",", @$coefs)));
}

#NOTA: Esta funcion deberia reescribirse tomando los datos del '.meta' porque si no, hay problemas con la represetacion del 0 (que es un archivo vacio)
sub getFormParams {
	my ($file) = @_;
	tie %F, "DB_File", $file or die "Cannot open file $file: $!\n";
	my @k = keys(%F);
	my ($diffs, $coefStr) = split(";", $k[0]);
	my @coefs = split(",", $coefStr);
	my $d = 0;
	map {$d += $_} @coefs;
	my %params;
	@params{n,r,d} = (scalar(@coefs),scalar(split(",", $diffs)),$d);
	return %params;
}

sub wedge {
	my ($fileW,$fileH) = @_;	
	my (%W, %H, %WEDGE);

	my %w = getFormParams($fileW);
	tie %W, "DB_File", $fileW or die "Cannot open file $fileW: $!\n";

	my %h = getFormParams($fileH);
	tie %H, "DB_File", $fileH or die "Cannot open file $fileH: $!\n";
# print "W=$fileW (alpha, dw) H=$fileH (beta, dh)\n";
	my $fileWedge = formCoefficients($h{n},$w{r} + $h{r},$w{d} + $h{d}, $fileW . "^" . $fileH);
	tie %WEDGE, "DB_File", $fileWedge or die "Cannot open file $fileWedge: $!\n";

	map {
		my (%dwSet, %alphaSet);
		my $kGammaKey = $_;
		my ($kStr, $gammaStr) = split(";", $kGammaKey);
		
		# calculamos dw (los diferenciales de 'w' que son <= que los diferenciales de 'dk')	
		my @dk = split(",", $kStr);
		my $dwFile = "tmp/" . $w{n} . "_" . $w{r} . "_" . $kStr . ".nr";
# print "INICIO: " . $dwFile . " n=" . $w{n} ." r=" . $w{r} . " list=" . join(",", @dk) . "\n"; 
		if (! -e $dwFile) {
			tie %dwSet, "DB_File", $dwFile or die "Cannot open file $dwFile: $!\n";
			combN_R(\%dwSet, 1, scalar(@dk), $w{r}, $w{r}, @dk);
		} else {
			tie %dwSet, "DB_File", $dwFile or die "Cannot open file $dwFile: $!\n";
		}
# print "FIN\n";
		# claculamos alpha (los monomios de 'w' que son <= que gamma: los monomios de k)
		my @gamma = split(",", $gammaStr);
		my $alphaFile = "tmp/" . $w{n} . "_" . $w{d} . "_" . $gammaStr . ".nd";
		if (! -e $alphaFile) {
			tie %alphaSet, "DB_File", $alphaFile or die "Cannot open file $alphaFile: $!\n"; 
			monomialsDegrees(\%alphaSet, 1, $w{n}, $w{d}, @gamma);
		} else {
			tie %alphaSet, "DB_File", $alphaFile or die "Cannot open file $alphaFile: $!\n"; 
		}

#		$WEDGE{$kGammaKey} = ""; # inicializamos el termino que queremos calcular
		map {
			# resta de conjuntos de los diferenciales: 'dk - dw' -> 'dh'
			my @dw = split(",", $_);
# print "dk: " . join(",", @dk) . "\n";
# print "dw: " . join(",", @dw) . "\n";
			my %dhTemp;
			@dhTemp{@dk} = "";
			delete @dhTemp{@dw};
			my @dh = sort(keys(%dhTemp));
# print "dh: ". join(",", @dh) . "\n";
			# 'gamma - alpha' -> beta
			map {				
				my @alpha = split(",", $_);
# print "gamma: " . join(",", @gamma) . "\n";
# print "alpha: " . join(",", @alpha) . "\n";
				my @beta;
				for (my $i = 0; $i < @gamma; $i++) {
					$beta[$i] = $gamma[$i] - $alpha[$i];
				}
# print "beta: " . join(",", @beta) . "\n";

				# aca va la operacion !!!!!
# print "alphakey: " . formKey(\@dw,\@alpha) . "\n";
# print "betakey: " . formKey(\@dh,\@beta) . "\n";
				my $sign = sign(\@dw,\@dh);
				my $termW = $W{formKey(\@dw,\@alpha)};
				my $termH = $H{formKey(\@dh,\@beta)};
				if ($termW && $termH) {
					$WEDGE{$kGammaKey} .= (!$WEDGE{$kGammaKey} && $sign eq "+" ? "" : $sign ) .
								"(" . $termW . ")*(" . $termH . ")";
				}
			} keys(%alphaSet);

		} keys(%dwSet);
		
		untie %dwSet;
		untie %alphaSet;
		if (!$WEDGE{$kGammaKey}) {delete($WEDGE{$kGammaKey});}
	} keys(%WEDGE);
# print "keys W: " .join(" | ", keys(%W)) . "\n";
# print "keys H: " .join(" | ", keys(%H)) . "\n";
	untie %W;
	untie %H;
	untie %WEDGE;
	return $fileWedge;
}

sub sign {
	my @dw = @{$_[0]};
	my @dh = @{$_[1]};
	my ($i, $j, $s) = (0,0,0);
	while ($i < scalar(@dw) && $j < scalar(@dh)) {
		if ($dw[$i] < $dh[$j]) {
			$i++;
		} else {
			$j++;
			$s += scalar(@dw) - $i;
		}
	}
	return ($s % 2) == 0 ? "+" : "-";
}

#TODO: hacer campos simetricos
sub vectorField {
	my ($n, $e, $fileName, $varName) = @_;

	return formCoefficients($n, 1, $e, $fileName, $varName);
}

sub radialField {
	my ($n) = @_;
	my $fileR = vectorField($n, 1, "R_" . $n);
	tie %R, "DB_File", $fileR or die "Cannot open file $fileR: $!\n";
	map {
		my $kGammaKey = $_;
		my ($k, $gammaStr) = split(";", $kGammaKey);
		my @gamma = split(",", $gammaStr);
		$R{$kGammaKey} = $gamma[$k-1] == 1 ? 1 : 0;
	} keys(%R);
	untie %R;
	return $fileR;
}

sub contraction {
	my ($fileX,$fileW) = @_;
	my (%W, %X, %CONTRACTION);

	my %w = getFormParams($fileW);
	tie %W, "DB_File", $fileW or die "Cannot open file $fileW: $!\n";

	my %x = getFormParams($fileX);
	tie %X, "DB_File", $fileX or die "Cannot open file $fileX: $!\n";
	
	my $fileContraction = formCoefficients($x{n},$w{r} - 1,$w{d} + $x{d}, $fileX . "L" . $fileW);
	tie %CONTRACTION, "DB_File", $fileContraction or die "Cannot open file $fileContraction: $!\n";

	map {
		my $kGammaKey = $_;
		my ($kStr, $gammaStr) = split(";", $kGammaKey);
		my @gamma = split(",", $gammaStr);
		my %alphaSet;
		my $alphaFile = "tmp/" . $x{n} . "_" . $x{d} . "_" . $gammaStr . ".nd";
		if (! -e $alphaFile) {
			tie %alphaSet, "DB_File", $alphaFile or die "Cannot open file $alphaFile: $!\n"; 
			monomialsDegrees(\%alphaSet, 1, $x{n}, $x{d}, @gamma);
		} else {
			tie %alphaSet, "DB_File", $alphaFile or die "Cannot open file $alphaFile: $!\n"; 
		}
#		$CONTRACTION{$kGammaKey} = "";
		my %dk;
		@dk{split(",", $kStr)} = "";
		for (my $i = 1; $i <= $x{n}; $i++) {
			if (! exists($dk{$i})) {
				my @dw = sort(keys(%dk),$i);
				map {				
					my @alpha = split(",", $_);
					my @beta;
					for (my $j = 0; $j < @gamma; $j++) {
						$beta[$j] = $gamma[$j] - $alpha[$j];
					}
					my $sign = sign([$i],[sort(keys(%dk))]);
					my $termX = $X{formKey([$i],\@alpha)};
					my $termW = $W{formKey(\@dw,\@beta)};
					if ($termX && $termW) {
						$CONTRACTION{$kGammaKey} .= (!$CONTRACTION{$kGammaKey} && $sign eq "+" ? "" : $sign ) .
									"(" . $termX  . ")*(" .  $termW . ")";
					}
				} keys(%alphaSet);
			}
		}
		untie %alphaSet;
		if (!$CONTRACTION{$kGammaKey}) {delete($CONTRACTION{$kGammaKey});}
	} keys(%CONTRACTION);
	untie %W;
	untie %X;
	untie %CONTRACTION;
	return $fileContraction;
}

sub diffExt {
	my ($fileW) = @_;
	my %w = getFormParams($fileW);
	tie %W, "DB_File", $fileW or die "Cannot open file $fileW: $!\n";
	my $fileDiffExt = formCoefficients($w{n},$w{r} + 1,$w{d} - 1, "d" . $fileW);
	tie %DIFF_EXT, "DB_File", $fileDiffExt or die "Cannot open file $fileDiffExt: $!\n";
	map {
		my $kGammaKey = $_;
		my ($kStr, $gammaStr) = split(";", $kGammaKey);
		my @gamma = split(",", $gammaStr);
		my %dk;
		@dk{split(",", $kStr)} = "";
		for (my $i = 1; $i <= $w{n}; $i++) {
			if (exists($dk{$i})) {
				my @gammaI = @gamma;
				$gammaI[$i-1]++;
				my %dkI = %dk;
				delete $dkI{$i};
				my @dkISorted = sort(keys(%dkI));
# print "i: $i kGammaKey: $kGammaKey dkISorted: " . join(",", @dkISorted) . " gammaI: " . join(",", @gammaI) . " formKey: " . formKey(\@dkISorted,\@gammaI) . "\n";
				my $sign = sign([$i],\@dkISorted);
				my $gammaTerm = $gammaI[$i-1];
				my $termW = $W{formKey(\@dkISorted,\@gammaI)};
				if ($gammaTerm && $termW) {
					$DIFF_EXT{$kGammaKey} .= (!$DIFF_EXT{$kGammaKey} && $sign eq "+" ? "" : $sign );
					$DIFF_EXT{$kGammaKey} .= $gammaTerm > 1 ? $gammaTerm  . "*(" .  $termW . ")" : "(" .  $termW . ")";
				}
# print "DIFF_EXT{...}: " . $DIFF_EXT{$kGammaKey} . "\n";
			}
		}
		if (!$DIFF_EXT{$kGammaKey}) {delete($DIFF_EXT{$kGammaKey});}
	} keys(%DIFF_EXT);
	untie %W;
	untie %DIFF_EXT;
	return $fileDiffExt;	
}

sub add {
	return operationForm(@_, "+");
}

sub subtract {
	return operationForm(@_, "-");
}

sub operationForm {
	my ($fileW, $fileH, $operation) = @_;
	
	my %w = getFormParams($fileW);
	tie %W, "DB_File", $fileW or die "Cannot open file $fileW: $!\n";

	my %h = getFormParams($fileH);
	tie %H, "DB_File", $fileH or die "Cannot open file $fileH: $!\n";		

	my $fileAdd = formCoefficients($w{n},$w{r},$w{d}, $fileW . $operation . $fileH);
	tie %ADD, "DB_File", $fileAdd or die "Cannot open file $fileAdd: $!\n";
	
	map {
		my $termW = $W{$_};
		my $termH = $H{$_};
		if ($termW) {
			$ADD{$_} = $termW;
		}
		if ($termH) {
			if ($ADD{$_}) {
				$ADD{$_} .= $operation . "(" . $termH . ")";
			} else {
				$ADD{$_} = $termH;
			}
		}
		if (!$ADD{$_}) {delete($ADD{$_});}
	} keys(%ADD);	
	untie %W;
	untie %H;
	untie %ADD;
	
	return $fileAdd;
}

# sub lieDerivative {
# 	my ($fileX, $fileW) = @_;
#	return add(contraction($fileX, diffExt($fileW)),diffExt(contraction($fileX, $fileW)));
#}

sub lieBracket {
	my ($fileW, $fileH) = @_;

	my (%W, %H, %BRACKET);

	my %w = getFormParams($fileW);
	tie %W, "DB_File", $fileW or die "Cannot open file $fileW: $!\n";

	my %h = getFormParams($fileH);
	tie %H, "DB_File", $fileH or die "Cannot open file $fileH: $!\n";
# print "W=$fileW (alpha, dw) H=$fileH (beta, dh)\n";
	my $fileBracket = formCoefficients($h{n},1,$w{d} + $h{d} - 1, "[" . $fileW . "," . $fileH . "]");
	tie %BRACKET, "DB_File", $fileBracket or die "Cannot open file $fileBracket: $!\n";

	my %alphaSet;
	my $alphaFile = "tmp/" . $w{n} . "_" . $w{d} . ".nd";
	if (! -e $alphaFile) {
		tie %alphaSet, "DB_File", $alphaFile or die "Cannot open file $alphaFile: $!\n"; 
		monomialsDegrees(\%alphaSet, 1, $w{n}, $w{d});
	} else {
		tie %alphaSet, "DB_File", $alphaFile or die "Cannot open file $alphaFile: $!\n"; 
	}

	my %betaSet;
	my $betaFile = "tmp/" . $h{n} . "_" . $h{d} . ".nd";
	if (! -e $betaFile) {
		tie %betaSet, "DB_File", $betaFile or die "Cannot open file $betaFile: $!\n"; 
		monomialsDegrees(\%betaSet, 1, $h{n}, $h{d});
	} else {
		tie %betaSet, "DB_File", $betaFile or die "Cannot open file $betaFile: $!\n"; 
	}

	my $n = $w{n};
	map {
		my @alpha = split(",", $_);
		map {
			my @beta = split(",", $_);
			for (my $i = 0; $i < $n; $i++) {
				for (my $j = 0; $j < $n; $j++) {
					my @gamma;
					for (my $k = 0; $k < $n; $k++) {
						$gamma[$k] = $alpha[$k] + $beta[$k];
					}
					if ($gamma[$j]>0) {
						$gamma[$j]--;
# print "i: $i j: $j kGammaKey: $kGammaKey iBeta: $iBeta\n";
						my $hKey = $H{formKey([$i+1],\@beta)};
						my $wKey = $W{formKey([$j+1],\@alpha)};
						my $kGammaKey = formKey([$i+1],\@gamma);
						if ($beta[$j] && $hKey && $wKey) {
							$BRACKET{$kGammaKey} .= ($BRACKET{$kGammaKey} ? "+" : "") . $beta[$j] . "*(" . $hKey . ")*(" . $wKey . ")";
						}

						$hKey = $H{formKey([$j+1],\@beta)};
						$wKey = $W{formKey([$i+1],\@alpha)};
						if ($alpha[$j] && $hKey && $wKey) {
							$BRACKET{$kGammaKey} .= "-" . $alpha[$j] . "*(" . $wKey . ")*(" . $hKey . ")";
						}
					}
				}
			}
		
		} keys(%betaSet);
	} keys(%alphaSet);
	
	map {if (!$BRACKET{$_}) {delete($BRACKET{$_});}} keys(%BRACKET);
	untie %W;
	untie %H;
	untie %BRACKET;
	return $fileBracket;

}

# polinomios
sub polynomialCoefficients {
	my ($n, $d, $coefFile, $varName) = @_;
	my $monomialsFile = "tmp/" . $n . "_" . $d . ".nd";
	if (! -e $monomialsFile) {
		my %h;
		tie %h, "DB_File", $monomialsFile or die "Cannot open file $monomialsFile: $!\n";
		monomialsDegrees(\%h, 1, $n, $d);
		untie %h;
	}
	if (-e $coefFile) {
		unlink($coefFile);
	}
	my (%FH_COEF, %FH_MONOM);
	tie %FH_COEF, "DB_File", $coefFile or die "Cannot open file $coefFile: $!\n";
	tie %FH_MONOM, "DB_File", $monomialsFile or die "Cannot open file 2: $!\n";
	map {
		my $value = $_;
		$FH_COEF{$value} = !$varName ? "" : $varName . "_" . $value;
	} keys(%FH_MONOM);
	untie %FH_COEF;
	untie %FH_MONOM;
	return $coefFile;	
}

sub diffPoly {
	my ($fileW) = @_;
	my %w = getPolyParams($fileW);
	tie %W, "DB_File", $fileW or die "Cannot open file $fileW: $!\n";
	my $fileDiffExt = formCoefficients($w{n},1,$w{d} - 1, "d" . $fileW);
	tie %DIFF_EXT, "DB_File", $fileDiffExt or die "Cannot open file $fileDiffExt: $!\n";
	map {
		my $kGammaKey = $_;
		my ($dk, $gammaStr) = split(";", $kGammaKey);
		my @gamma = split(",", $gammaStr);
		$gamma[$dk-1]++;

		my $gammaTerm = $gamma[$dk-1];
		my $termW = $W{join(",", @gamma)};
		if ($gammaTerm && $termW) {
			$DIFF_EXT{$kGammaKey} = $gammaTerm > 1 ? $gammaTerm  . "*(" .  $termW . ")" : $termW;	
		}
		if (!$DIFF_EXT{$kGammaKey}) {delete($DIFF_EXT{$kGammaKey});}
	} keys(%DIFF_EXT);
	untie %W;
	untie %DIFF_EXT;
	return $fileDiffExt;	
}

sub getPolyParams {
	my ($file) = @_;
	tie %F, "DB_File", $file or die "Cannot open file $file: $!\n";
	my @k = keys(%F);
	my $coefStr = $k[0];
	my @coefs = split(",", $coefStr);
	my $d = 0;
	map {$d += $_} @coefs;
	my %params;
	@params{n,d} = (scalar(@coefs),$d);
	return %params;
}

sub wedgeFormPoly {
	my ($fileW,$fileH) = @_;
	my (%W, %H, %WEDGE);

	my %w = getFormParams($fileW);
	tie %W, "DB_File", $fileW or die "Cannot open file $fileW: $!\n";

	my %h = getPolyParams($fileH);
	tie %H, "DB_File", $fileH or die "Cannot open file $fileH: $!\n";
	my $fileWedge = formCoefficients($h{n},$w{r},$w{d} + $h{d}, $fileW . "^" . $fileH);
	tie %WEDGE, "DB_File", $fileWedge or die "Cannot open file $fileWedge: $!\n";

	map {
		my $kGammaKey = $_;
		my ($kStr, $gammaStr) = split(";", $kGammaKey);
		my @dk = split(",", $kStr);
		my @gamma = split(",", $gammaStr);
		
		my %alphaSet;
		my $alphaFile = "tmp/" . $w{n} . "_" . $w{d} . "_" . $gammaStr . ".nd";
		if (! -e $alphaFile) {
			tie %alphaSet, "DB_File", $alphaFile or die "Cannot open file $alphaFile: $!\n"; 
			monomialsDegrees(\%alphaSet, 1, $w{n}, $w{d}, @gamma);
		} else {
			tie %alphaSet, "DB_File", $alphaFile or die "Cannot open file $alphaFile: $!\n"; 
		}

#		$WEDGE{$kGammaKey} = ""; # inicializamos el termino que queremos calcular
		map {				
			my @alpha = split(",", $_);
			my @beta;
			for (my $i = 0; $i < @gamma; $i++) {
				$beta[$i] = $gamma[$i] - $alpha[$i];
			}

			my $termW = $W{formKey(\@dk,\@alpha)};
			my $termH = $H{join(",", @beta)};
			if ($termW && $termH) {
				$WEDGE{$kGammaKey} .= (!$WEDGE{$kGammaKey} ? "" : "+" ) .
							"(" . $termW . ")*(" . $termH . ")";
			}
		} keys(%alphaSet);
		untie %alphaSet;
		if (!$WEDGE{$kGammaKey}) {delete($WEDGE{$kGammaKey});}
	} keys(%WEDGE);
	untie %W;
	untie %H;
	untie %WEDGE;
	return $fileWedge;
}

sub wedgePolyPoly {
	my ($fileW,$fileH) = @_;
	my (%W, %H, %WEDGE);

	my %w = getPolyParams($fileW);
	tie %W, "DB_File", $fileW or die "Cannot open file $fileW: $!\n";
	my %h = getPolyParams($fileH);
	tie %H, "DB_File", $fileH or die "Cannot open file $fileH: $!\n";
	my $fileWedge = polynomialCoefficients($h{n},$w{d} + $h{d}, $fileW . "^" . $fileH);
	tie %WEDGE, "DB_File", $fileWedge or die "Cannot open file $fileWedge: $!\n";

	map {
		my $gammaStr = $_;
		my @gamma = split(",", $gammaStr);
		
		my %alphaSet;
		my $alphaFile = "tmp/" . $w{n} . "_" . $w{d} . "_" . $gammaStr . ".nd";
		if (! -e $alphaFile) {
			tie %alphaSet, "DB_File", $alphaFile or die "Cannot open file $alphaFile: $!\n"; 
			monomialsDegrees(\%alphaSet, 1, $w{n}, $w{d}, @gamma);
		} else {
			tie %alphaSet, "DB_File", $alphaFile or die "Cannot open file $alphaFile: $!\n"; 
		}

#		$WEDGE{$gammaStr} = ""; # inicializamos el termino que queremos calcular
		map {
			my @alpha = split(",", $_);
			my @beta;
			for (my $i = 0; $i < @gamma; $i++) {
				$beta[$i] = $gamma[$i] - $alpha[$i];
			}

			my $termW = $W{join(",", @alpha)};
			my $termH = $H{join(",", @beta)};
			if ($termW && $termH) {
				$WEDGE{$gammaStr} .= (!$WEDGE{$gammaStr} ? "" : "+" ) .
							"(" . $termW . ")*(" . $termH . ")";
			}
		} keys(%alphaSet);
		untie %alphaSet;
		if (!$WEDGE{$gammaStr}) {delete($WEDGE{$gammaStr});}
	} keys(%WEDGE);
	untie %W;
	untie %H;
	untie %WEDGE;
	return $fileWedge;
}

sub addPoly {
	return operationPoly(@_, "+");
}

sub subtractPoly {
	return operationPoly(@_, "-");
}

sub operationPoly {
	my ($fileW, $fileH, $operation) = @_;
	
	my %w = getPolyParams($fileW);
	tie %W, "DB_File", $fileW or die "Cannot open file $fileW: $!\n";

	my %h = getPolyParams($fileH);
	tie %H, "DB_File", $fileH or die "Cannot open file $fileH: $!\n";		

	my $fileAdd = polynomialCoefficients($w{n},$w{d}, $fileW . $operation . $fileH);
	tie %ADD, "DB_File", $fileAdd or die "Cannot open file $fileAdd: $!\n";
	
	map {
		my $termW = $W{$_};
		my $termH = $H{$_};
		if ($termW) {
			$ADD{$_} = $termW;
		}
		if ($termH) {
			if ($ADD{$_}) {
				$ADD{$_} .= $operation . "(" . $termH . ")";
			} else {
				$ADD{$_} = $termH;
			}
		}
		if (!$ADD{$_}) {delete($ADD{$_});}
	} keys(%ADD);	
	untie %W;
	untie %H;
	untie %ADD;
	
	return $fileAdd;
}

#contraer contra una 1-forma
sub contractionOne {
	my ($fileX,$fileW) = @_;
	my (%W, %X, %CONTRACTION);

	my %w = getFormParams($fileW);
	tie %W, "DB_File", $fileW or die "Cannot open file $fileW: $!\n";

	my %x = getFormParams($fileX);
	tie %X, "DB_File", $fileX or die "Cannot open file $fileX: $!\n";
	
	my $fileContraction = polynomialCoefficients($x{n},$w{d} + $x{d}, $fileX . "L" . $fileW);
	tie %CONTRACTION, "DB_File", $fileContraction or die "Cannot open file $fileContraction: $!\n";

	map {
		my $gammaStr = $_;
		my @gamma = split(",", $gammaStr);
		my %alphaSet;
		my $alphaFile = "tmp/" . $x{n} . "_" . $x{d} . "_" . $gammaStr . ".nd";
		if (! -e $alphaFile) {
			tie %alphaSet, "DB_File", $alphaFile or die "Cannot open file $alphaFile: $!\n"; 
			monomialsDegrees(\%alphaSet, 1, $x{n}, $x{d}, @gamma);
		} else {
			tie %alphaSet, "DB_File", $alphaFile or die "Cannot open file $alphaFile: $!\n"; 
		}
#		$CONTRACTION{$gammaStr} = "";
		for (my $i = 1; $i <= $x{n}; $i++) {
			map {
				my @alpha = split(",", $_);
				my @beta;
				for (my $j = 0; $j < @gamma; $j++) {
					$beta[$j] = $gamma[$j] - $alpha[$j];
				}

				my $termX = $X{formKey([$i],\@alpha)};
				my $termW = $W{formKey([$i],\@beta)};
				if ($termX && $termW) {
					$CONTRACTION{$gammaStr} .= (!$CONTRACTION{$gammaStr} ? "" : "+" ) .
								"(" . $termX  . ")*(" .  $termW . ")";
				}
			} keys(%alphaSet);
		}
		untie %alphaSet;
		if (!$CONTRACTION{$gammaStr}) {delete($CONTRACTION{$gammaStr});}
	} keys(%CONTRACTION);
	untie %W;
	untie %X;
	untie %CONTRACTION;
	return $fileContraction;
}

sub wedgeScalar {
	my ($num, $fileW) = @_;
	my $fileName = $num . "^" . $fileW;
	if (-e $fileName) {
		unlink($fileName);
	}
	tie %WEDGE, "DB_File", $fileName or die "Cannot open file $fileName : $!\n";
	tie %W, "DB_File", $fileW or die "Cannot open file $fileW: $!\n";
	
	map {
		if ($num && $W{$_}) {
			$WEDGE{$_} = "(" . $num . ")*(" . $W{$_} . ")";
		}
		if (!$WEDGE{$_}) {delete($WEDGE{$_});}
	} keys(%W);
	untie %W;
	untie %WEDGE;
	return $fileName;
}

1
