package Form;

# # diffAlg - Differential algebra library
# # 
# # Copyright (C) 2012 Manuel Dubinsky, Cesar Massari, Ariel Molinuevo
# # 
# # 
# # This program is free software: you can redistribute it and/or modify
# # it under the terms of the GNU General Public License as published by
# # the Free Software Foundation, either version 3 of the License, or
# # (at your option) any later version.
# # This program is distributed in the hope that it will be useful,
# # but WITHOUT ANY WARRANTY; without even the implied warranty of
# # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# # GNU General Public License for more details.
# # You should have received a copy of the GNU General Public License
# # along with this program. If not, see <http://www.gnu.org/licenses/>.

use DiffAlgebra;
use Exporter;

@ISA = qw(Exporter);

@EXPORT = qw(newForm add subtract wedge wedgeScalar contraction vectorField d R lieBracket);

# Operadores (los campos son: Asociatividad, Aridad, Nombre)
# None 	0 	Terms, and list operators (leftward)
# Left 	2 	->
# None 	1 	++ --
# Right 	2 	**
# Right 	1 	! ~ > and unary + and -
# Left 	2 	=~ !~
# Left 	2 	* / % x
# Left 	2 	+ - .
# Left 	2 	<< >>
# Right 	0,1 	Named unary operators
# None 	2 	< > <= >= lt gt le ge
# None 	2 	== != <=> eq ne cmp
# Left 	2 	&
# Left 	2 	| ^
# Left 	2 	&&
# Left 	2 	||
# None 	2 	.. ...
# Right 	3 	?:
# Right 	2 	= += -= *= and so on
# Left 	2 	, =>
# Right 	0+ 	List operators (rightward)
# Right 	1 	not
# Left 	2 	and
# Left 	2 	or xor

use overload ('+'   => 'add',
		'-'   => 'subtract',
	      '^' => 'wedge',
	      '""' => 'printForm',
	      '||' => 'contraction');

sub newForm {
	my $self  = {};	
	my ($formFile, $formVars, $n, $r, $d, $varName);
	if (scalar(@_) == 5) {
		($n, $r, $d, $formFile, $varName) = @_;
		$formFile = $r > 0 ?
			DiffAlgebra::formCoefficients($n, $r, $d, $formFile, $varName) :
				DiffAlgebra::polynomialCoefficients($n, $d, $formFile, $varName);
		tie %h, "DB_File", $formFile or die "Cannot open file " . $formFile . ": $!\n";
		$formVars = join(" ",values(%h));
		untie %h;
	} elsif (scalar(@_) == 6) {
		($n, $r, $d, $formFile, $formVars) = @_;
	} elsif (scalar(@_) == 1) {
		my $file = shift;
		tie %h, "DB_File", $file . ".meta" or die "Cannot open file " . $file . ".meta: $!\n";
		($formFile, $formVars, $n, $r, $d) = ($h{formFile}, $h{formVars}, $h{n}, $h{r}, $h{d});
		untie %h;
	}
	bless($self);
	$self->setParams($formFile, $formVars, $n, $r, $d);
	return $self;
}

sub setParams {
	my ($f, $formFile, $formVars, $n, $r, $d) = @_;
	$f->{formFile} = $formFile;
	$f->{formVars} = $formVars;
	$f->{n} = $n;
	$f->{r} = $r;
	$f->{d} = $d;
	$f->save();
	return $f;
}

sub printForm {
	my ($f) = @_;
	return "n: " . $f->{n} . " r: " . $f->{r} . " d: " . $f->{d} . "\nfile: " . $f->{formFile} . " vars: " . $f->{formVars} . "\n";
}

sub isPoly {
	my $f = shift;
	return $f->{r} == 0;
}

sub isZero {
	my ($f1) = @_;
	return $f1->{formVars} eq "0";
}

sub makeZero {
	my ($formFile) = @_;
	return newForm(0, 0, 0, $formFile, "0", "INTERNAL");
}

sub add {
	my ($f1, $f2) = @_;
	if (isZero($f1) || isZero($f2)) {
		return makeZero($f1->{formFile} . "+" . $f2->{formFile});
	}
	if ($f1->{n} == $f2->{n} && $f1->{r} == $f2->{r} && $f1->{d} == $f2->{d}) { # esto es Ariel
		my $fileAdd;
		if (!$f1->isPoly()) {
			$fileAdd = DiffAlgebra::add($f1->{formFile},$f2->{formFile});
		} else {
			$fileAdd = DiffAlgebra::addPoly($f1->{formFile},$f2->{formFile});
		}
		return newForm($f1->{n}, $f1->{r}, $f1->{d}, $fileAdd, $f1->{formVars} . " " . $f2->{formVars}, "INTERNAL");
	}
	die "ADD Bad Command";
}

sub subtract {
	my ($f1, $f2) = @_;
	if (isZero($f1) || isZero($f2)) {
		return makeZero($f1->{formFile} . "-" . $f2->{formFile});
	}
	if ($f1->{n} == $f2->{n} && $f1->{r} == $f2->{r} && $f1->{d} == $f2->{d}) { # esto es Ariel
		my $fileAdd;
		if (!$f1->isPoly()) {
			$fileAdd = DiffAlgebra::subtract($f1->{formFile},$f2->{formFile});
		} else {
			$fileAdd = DiffAlgebra::subtractPoly($f1->{formFile},$f2->{formFile});
		}
		return newForm($f1->{n}, $f1->{r}, $f1->{d}, $fileAdd, $f1->{formVars} . " " . $f2->{formVars}, "INTERNAL");
	}
	die "SUBTRACT Bad Command";
}

sub wedge {
	my ($f1, $f2) = @_;
	if (isZero($f1) || isZero($f2)) {
		return isZero($f1) ? $f1 : $f2;
	}
	if ($f1->{n} == $f2->{n}) {
		my $fileWedge;
		if ($f1->{r} + $f2->{r} <= $f1->{n}) {
			if ($f1->{r} * $f2->{r} > 0) {
				$fileWedge = DiffAlgebra::wedge($f1->{formFile},$f2->{formFile});
			} elsif ($f1->{r} + $f2->{r} > 0) {
				if ($f2->isPoly()) {
					$fileWedge = DiffAlgebra::wedgeFormPoly($f1->{formFile},$f2->{formFile});				
				} else {
					$fileWedge = DiffAlgebra::wedgeFormPoly($f2->{formFile},$f1->{formFile});
				}
			} else {
				$fileWedge = DiffAlgebra::wedgePolyPoly($f1->{formFile},$f2->{formFile});
			}
		} else {
			return makeZero($f1->{formFile} . "^" . $f2->{formFile});
		}
		return newForm($f1->{n}, $f1->{r} + $f2->{r}, $f1->{d} + $f2->{d}, $fileWedge, $f1->{formVars} . " " . $f2->{formVars}, "INTERNAL");
	}
	die "WEDGE Different n form 1:" . $f1->{formFile} . " form 2: " . $f2->{formFile};
}

sub wedgeScalar {
	my ($num, $f1) = @_;
	if (!$num || isZero($f1)) {
		return makeZero($num . "*" . $f1->{formFile});
	}
	
	my $fileMult = DiffAlgebra::wedgeScalar( $num, $f1->{formFile} );
	return newForm($f1->{n}, $f1->{r}, $f1->{d}, $fileMult, $f1->{formVars}, "INTERNAL");
}
	
sub diffExt {
	my ($f1) = @_;
	if (isZero($f1) || ! $f1->{d} || ($f1->{r} >= $f1->{n})) {
		return makeZero("d" . $f1->{formFile});
	}
	my $fileDiff;
	if (!$f1->isPoly()) {
		$fileDiff = DiffAlgebra::diffExt($f1->{formFile});
	} else {
		$fileDiff = DiffAlgebra::diffPoly($f1->{formFile});
	}
	return newForm($f1->{n}, $f1->{r} + 1, $f1->{d} - 1, $fileDiff, $f1->{formVars}, "INTERNAL"); 
}

sub d {
	return diffExt(@_);
}

sub contraction {
	my ($f1, $f2) = @_;	
	if (isZero($f1) || isZero($f2) || $f2->isPoly()) {
		return makeZero($f1->{formFile} . "L" . $f2->{formFile});
	}
	my $fileC;
	if ($f1->{r} !=1) {
		die "CONTRACTION Error";
	} elsif ($f2->{r} > 1) {
		$fileC = DiffAlgebra::contraction($f1->{formFile},$f2->{formFile});		
	} else {
		$fileC = DiffAlgebra::contractionOne($f1->{formFile},$f2->{formFile});
	}
	return newForm($f2->{n}, $f2->{r} - 1, $f1->{d} + $f2->{d}, $fileC, $f1->{formVars} . " " . $f2->{formVars}, "INTERNAL");
}

sub radialField {
	my ($n) = @_;
	return newForm($n, 1, 1, DiffAlgebra::radialField($n), "", "INTERNAL");
}

sub R {
	return radialField(@_);
}

sub vectorField {
	my ($n, $e, $fileName, $varName) = @_;
	return newForm($n, 1, $e, $fileName, $varName);
}

sub lieBracket {
	my ($f1, $f2) = @_;
	if (isZero($f1) || isZero($f2) || ($f1->{d} + $f2->{d}) == 0) {
		return makeZero("[" . $f1->{formFile} . "," . $f2->{formFile} . "]");
	}
	if ($f1->{r} !=1 || $f2->{r} !=1 || $f1->{n} != $f2->{n}) {
		die "LIE BRACKET Error";
	}
	my $fileC = DiffAlgebra::lieBracket($f1->{formFile},$f2->{formFile});
	return newForm($f2->{n}, 1, $f1->{d} + $f2->{d} - 1, $fileC, $f1->{formVars} . " " . $f2->{formVars}, "INTERNAL");	
}

sub save {
	my $self = shift;
	my $file = $self->{formFile} . ".meta";
	if (-e $file) {
		unlink($file);
	}
	tie %h, "DB_File", $file;
	my %cleanedVars;
	@cleanedVars{split(" ", $self->{formVars})} = 0;
	$h{formFile} = $self->{formFile};
	$h{formVars} = join(" ", keys(%cleanedVars));
	$h{n} = $self->{n};
	$h{r} = $self->{r};
	$h{d} = $self->{d};
	untie %h;
}

1
