#!/usr/bin/perl

# # diffAlg - Differential algebra library
# # 
# # Copyright (C) 2012 Manuel Dubinsky, Cesar Massari, Ariel Molinuevo
# # 
# # 
# # This program is free software: you can redistribute it and/or modify
# # it under the terms of the GNU General Public License as published by
# # the Free Software Foundation, either version 3 of the License, or
# # (at your option) any later version.
# # This program is distributed in the hope that it will be useful,
# # but WITHOUT ANY WARRANTY; without even the implied warranty of
# # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# # GNU General Public License for more details.
# # You should have received a copy of the GNU General Public License
# # along with this program. If not, see <http://www.gnu.org/licenses/>.

use Form;
require "formrc";

my @c = ("ls", "newForm", "add", "subtract", "wedge", "wedgeScalar", "contraction", "vectorField", "d", "R", "lieBracket", "toSingular");

sub help {
	for (my $i = 0; $i < scalar(@c); $i++) {
		print "$i - $c[$i]\n";
	}
}

sub inputParams {
	my @params = @_;
	my %h;
	for (my $i = 0; $i < scalar(@params); $i++) {
		print "$params[$i] ?: ";
		$h{$params[$i]} = <STDIN>;
		chomp($h{$params[$i]});
	}
	return %h;
}

sub ls {
	my @p = `ls -1 *.meta`;
	for (my $i = 0; $i < scalar(@p); $i++) {
		$p[$i] =~ s/\.meta//g;
		print "$p[$i]";
	}
}

sub operation {
	my @paramNames = @_;
	my @p = `ls -1 *.meta`;
	for (my $i = 0; $i < scalar(@p); $i++) {
		chomp($p[$i]);
		$p[$i] =~ s/\.meta//g;
		print "$i - $p[$i]\n";
	}
	my %h = inputParams(@paramNames);
	map { $h{$_} = $p[$h{$_}];} keys(%h);
	return %h;
}

sub newForm {
	my %h = inputParams("n", "r", "d", "formName", "varName");
	Form::newForm($h{n},$h{r},$h{d},$h{formName},$h{varName});
}

sub add {
	my %h = operation("form_1", "form_2");
	Form::newForm($h{form_1}) + Form::newForm($h{form_2});
}

sub subtract {
	my %h = operation("form_1", "form_2");
	Form::newForm($h{form_1}) - Form::newForm($h{form_2});
}

sub wedge {
	my %h = operation("form_1", "form_2");
	Form::newForm($h{form_1}) ^ Form::newForm($h{form_2});
}

sub contraction {
	my %h = operation("field", "form");
	Form::contraction(Form::newForm($h{field}), Form::newForm($h{form}));
}

sub d {
	my %h = operation("form");
	Form::d(Form::newForm($h{form}));
}

sub vectorField {
	my %h = inputParams("n", "e", "fieldName", "varName");
	Form::vectorField($h{n},$h{e},$h{fieldName},$h{varName});
}

sub R {
	my %h = inputParams("n");
	Form::R($h{n});
}

sub lieBracket {
	my %h = operation("field_1", "field_2");
	Form::lieBracket(Form::newForm($h{field_1}), Form::newForm($h{field_2}));
}

sub wedgeScalar {
	my %h = operation("file");
	my %g = inputParams("scalar");
	Form::wedgeScalar($g{scalar}, Form::newForm($h{file}));
}

sub toSingular {
	my %h = operation("file");
	`./toSingular.pl $h{file}`;
}

my $prompt = "> ";
print "diffAlg - Differential algebra library\nCopyright (C) 2012 Manuel Dubinsky, Cesar Massari, Ariel Molinuevo\nThis program comes with ABSOLUTELY NO WARRANTY.\nThis is free software, and you are welcome to redistribute it under certain conditions.\n\n";
print "Type 'help' or 'exit'\n" . $prompt;
while (<STDIN>) {
	chomp;
	if ($_ eq 'exit') {
		print "Hasta la proxima Fede...\n";
		break;
	}
	if ($_ =~ /^\!\s*(.*)/) {
		print `$1`;
	} elsif ($_ =~ /^(\d+)/) {
		if ($1 < scalar(@c)) {
			&{$c[$1]};
		}
	} else {
		eval($_);
	}
	print $prompt;
}
